import re
from typing import (
    Annotated,
    Callable,
    get_origin,
    TYPE_CHECKING,
    Union,
)

from pydantic import (
    BeforeValidator,
    Field,
    PlainSerializer,
    WithJsonSchema,
)
from pydantic_core import PydanticCustomError
from typing_extensions import (
    get_args,
)

from galaxy.exceptions import MessageException

if TYPE_CHECKING:
    from galaxy.security.idencoding import IdEncodingHelper

ENCODED_DATABASE_ID_PATTERN = re.compile("f?[0-9a-f]+")
ENCODED_ID_LENGTH_MULTIPLE = 16


def validation_message_wrapper(callable: Callable):
    """Wraps MessageException in a PydanticCustomError."""

    def wrapper(database_id):
        try:
            return callable(database_id)
        except MessageException as e:
            # we want to return it as a PydanticCustomError
            # so that it can be handled by the Pydantic error handling system, so we can restore the
            # original exception.
            raise PydanticCustomError("message_exception", "A seralizable exception occurrred", {"exception": e})

    return wrapper


@validation_message_wrapper
def encode_id(database_id: int) -> str:
    return Security.security.encode_id(database_id)


@validation_message_wrapper
def decode_id(encoded_id: str) -> int:
    return Security.security.decode_id(encoded_id)


class Security:
    security: "IdEncodingHelper"


def ensure_valid_id(v: str) -> str:
    assert isinstance(v, str)
    len_v = len(v)
    if len_v % ENCODED_ID_LENGTH_MULTIPLE:
        raise ValueError("Invalid id length, must be multiple of 16")
    m = ENCODED_DATABASE_ID_PATTERN.fullmatch(v.lower())
    if not m:
        raise ValueError("Invalid characters in encoded ID")
    return v


def ensure_valid_folder_id(v):
    if not isinstance(v, str):
        raise ValueError("String required")
    if not v.startswith("F"):
        raise ValueError("Invalid library folder ID. Folder IDs must start with an 'F'")
    v = v[1:]
    ensure_valid_id(v)
    return v


DecodedDatabaseIdField = Annotated[
    int,
    BeforeValidator(decode_id),
    BeforeValidator(ensure_valid_id),
    PlainSerializer(encode_id, return_type=str, when_used="json"),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="serialization",
    ),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="validation",
    ),
]

EncodedDatabaseIdField = Annotated[
    str,
    BeforeValidator(encode_id),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="serialization",
    ),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="validation",
    ),
]

LibraryFolderDatabaseIdField = Annotated[
    int,
    BeforeValidator(decode_id),
    BeforeValidator(ensure_valid_folder_id),
    PlainSerializer(
        lambda database_id: f"F{Security.security.encode_id(database_id)}", return_type=str, when_used="json"
    ),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="serialization",
    ),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="validation",
    ),
]

EncodedLibraryFolderDatabaseIdField = Annotated[
    str,
    BeforeValidator(lambda database_id: f"F{encode_id(database_id)}"),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="serialization",
    ),
    WithJsonSchema(
        {"type": "string", "example": "0123456789ABCDEF", "pattern": "[0-9a-fA-F]+", "minLength": 16},
        mode="validation",
    ),
]


def literal_to_value(arg):
    val = get_args(arg)
    if not val:
        return arg
    if len(val) > 1:
        raise Exception("Can't extract default argument for unions")
    return val[0]


def is_optional(field):
    args = get_args(field)
    return get_origin(field) is Union and len(args) == 2 and type(None) in args


def ModelClassField(default_value):
    """Represents a database model class name annotated as a constant
    pydantic Field.
    :param class_name: The name of the database class.
    :return: A constant pydantic Field with default annotations for model classes.
    """
    return Field(
        ...,
        title="Model class",
        description="The name of the database model class.",
        json_schema_extra={"const": literal_to_value(default_value), "type": "string"},
    )


def accept_wildcard_defaults_to_json(v):
    assert isinstance(v, str)
    # Accept header can have multiple comma separated values.
    # If any of these values is the wildcard - we default to application/json.
    if "*/*" in v:
        return "application/json"
    return v


AcceptHeaderValidator = BeforeValidator(accept_wildcard_defaults_to_json)



# generated by datamodel-codegen:
#   filename:  https://opensource.ieee.org/2791-object/ieee-2791-schema/-/raw/master/description_domain.json
#   timestamp: 2022-09-13T23:51:48+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import (
    List,
    Optional,
)

from pydantic import (
    AnyUrl,
    BaseModel,
    ConfigDict,
    EmailStr,
    Field,
    RootModel,
)


class XrefItem(BaseModel):
    namespace: str = Field(..., description="External resource vendor prefix", examples=["pubchem.compound"])
    name: str = Field(..., description="Name of external reference", examples=["PubChem-compound"])
    ids: List[str] = Field(..., description="List of reference identifiers")
    access_time: datetime = Field(..., description="Date and time the external reference was accessed")


class Uri(BaseModel):
    model_config = ConfigDict(extra="forbid")

    filename: Optional[str] = None
    uri: AnyUrl
    access_time: Optional[datetime] = Field(
        None, description="Time stamp of when the request for this data was submitted"
    )
    sha1_checksum: Optional[str] = Field(
        None, description="output of hash function that produces a message digest", pattern="[A-Za-z0-9]+"
    )


class ObjectId(RootModel):
    root: str = Field(
        ...,
        description="A unique identifier that should be applied to each IEEE-2791 Object instance, generated and assigned by a IEEE-2791 database engine. IDs should never be reused",
    )


class ContributionEnum(Enum):
    authoredBy = "authoredBy"
    contributedBy = "contributedBy"
    createdAt = "createdAt"
    createdBy = "createdBy"
    createdWith = "createdWith"
    curatedBy = "curatedBy"
    derivedFrom = "derivedFrom"
    importedBy = "importedBy"
    importedFrom = "importedFrom"
    providedBy = "providedBy"
    retrievedBy = "retrievedBy"
    retrievedFrom = "retrievedFrom"
    sourceAccessedBy = "sourceAccessedBy"


class Contributor(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str = Field(..., description="Name of contributor", examples=["Charles Darwin"])
    affiliation: Optional[str] = Field(
        None, description="Organization the particular contributor is affiliated with", examples=["HMS Beagle"]
    )
    email: Optional[EmailStr] = Field(
        None,
        description="electronic means for identification and communication purposes",
        examples=["name@example.edu"],
    )
    contribution: List[ContributionEnum] = Field(
        ..., description="type of contribution determined according to PAV ontology"
    )
    orcid: Optional[AnyUrl] = Field(
        None,
        description="Field to record author information. ORCID identifiers allow for the author to curate their information after submission. ORCID identifiers must be valid and must have the prefix ‘https://orcid.org/’",
        examples=["http://orcid.org/0000-0002-1825-0097"],
    )


class PrerequisiteItem(BaseModel):
    name: str = Field(
        ..., description="Public searchable name for reference or prereq.", examples=["Hepatitis C virus genotype 1"]
    )
    uri: Uri


class PipelineStep(BaseModel):
    model_config = ConfigDict(extra="forbid")

    step_number: int = Field(
        ...,
        description="Non-negative integer value representing the position of the tool in a one-dimensional representation of the pipeline.",
    )
    name: str = Field(..., description="This is a recognized name of the software tool", examples=["HIVE-hexagon"])
    description: str = Field(
        ..., description="Specific purpose of the tool.", examples=["Alignment of reads to a set of references"]
    )
    version: Optional[str] = Field(
        None,
        description="Version assigned to the instance of the tool used corresponding to the upstream release.",
        examples=["1.3"],
    )
    prerequisite: Optional[List[PrerequisiteItem]] = Field(None, description="Reference or required prereqs")
    input_list: List[Uri] = Field(..., description="URIs (expressed as a URN or URL) of the input files for each tool.")
    output_list: List[Uri] = Field(
        ..., description="URIs (expressed as a URN or URL) of the output files for each tool."
    )


class DescriptionDomain(BaseModel):
    keywords: List[str] = Field(..., description="Keywords to aid in search-ability and description of the object.")
    xref: Optional[List[XrefItem]] = Field(
        None, description="List of the databases or ontology IDs that are cross-referenced in the IEEE-2791 Object."
    )
    platform: Optional[List[str]] = Field(
        None,
        description="reference to a particular deployment of an existing platform where this IEEE-2791 Object can be reproduced.",
    )
    pipeline_steps: List[PipelineStep] = Field(
        ...,
        description="Each individual tool (or a well defined and reusable script) is represented as a step. Parallel processes are given the same step number.",
    )



from datetime import datetime
from enum import Enum
from typing import (
    Annotated,
    Any,
    Generic,
    Optional,
    Union,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    UUID1,
    UUID4,
)
from typing_extensions import (
    Literal,
    TypeAliasType,
)

from galaxy.schema import schema
from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
    literal_to_value,
    ModelClassField,
)
from galaxy.schema.generics import (
    DatabaseIdT,
    GenericModel,
)
from galaxy.schema.schema import (
    CreateTimeField,
    DataItemSourceType,
    IMPLICIT_COLLECTION_JOBS_MODEL_CLASS,
    INVOCATION_MODEL_CLASS,
    INVOCATION_STEP_MODEL_CLASS,
    JOB_MODEL_CLASS,
    JobState,
    Model,
    StoreContentSource,
    UpdateTimeField,
    WithModelClass,
)
from .workflows import (
    INPUTS_BY_DESCRIPTION,
    PreferredIntermediateObjectStoreIdField,
    PreferredObjectStoreIdField,
    PreferredOutputsObjectStoreIdField,
    ReplacementParametersField,
    ResourceParametersField,
    STEP_PARAMETERS_DESCRIPTION,
    STEP_PARAMETERS_NORMALIZED_DESCRIPTION,
    STEP_PARAMETERS_NORMALIZED_TITLE,
    STEP_PARAMETERS_TITLE,
    UseCachedJobField,
)

INVOCATION_STEP_OUTPUT_SRC = Literal["hda"]
INVOCATION_STEP_COLLECTION_OUTPUT_SRC = Literal["hdca"]
REPORT_RENDER_FORMAT_MARKDOWN = Literal["markdown"]


InvocationStepActionField = Field(
    title="Action",
    description="Whether to take action on the invocation step.",
)

InvocationIdField: EncodedDatabaseIdField = Field(
    default=...,
    title="ID",
    description="The encoded ID of the workflow invocation.",
)


class WarningReason(str, Enum):
    workflow_output_not_found = "workflow_output_not_found"


class FailureReason(str, Enum):
    dataset_failed = "dataset_failed"
    collection_failed = "collection_failed"
    job_failed = "job_failed"
    output_not_found = "output_not_found"
    expression_evaluation_failed = "expression_evaluation_failed"
    when_not_boolean = "when_not_boolean"
    unexpected_failure = "unexpected_failure"
    workflow_parameter_invalid = "workflow_parameter_invalid"


# The reasons below are attached to the invocation and user-actionable.
# Not included are `unexpected_failure` and `expression_evaluation_failed`.
# If expression evaluation fails we're not attaching the templated
# expression to the invocation, as it could contain secrets.
# If the failure reason is not in `FAILURE_REASONS_EXPECTED` we should
# log an exception so admins can debug and/or submit bug reports.
FAILURE_REASONS_EXPECTED = (
    FailureReason.dataset_failed,
    FailureReason.collection_failed,
    FailureReason.job_failed,
    FailureReason.output_not_found,
    FailureReason.when_not_boolean,
)


class CancelReason(str, Enum):
    """Possible reasons for a cancelled workflow."""

    history_deleted = "history_deleted"
    user_request = "user_request"
    cancelled_on_review = "cancelled_on_review"


class InvocationMessageBase(GenericModel):
    reason: Union[CancelReason, FailureReason, WarningReason]
    model_config = ConfigDict(from_attributes=True, populate_by_name=True)


class GenericInvocationCancellationReviewFailed(InvocationMessageBase, Generic[DatabaseIdT]):
    reason: Literal[CancelReason.cancelled_on_review]
    workflow_step_id: int = Field(
        ...,
        description="Workflow step id of paused step that did not pass review.",
        validation_alias="workflow_step_index",
    )


class GenericInvocationCancellationHistoryDeleted(InvocationMessageBase, Generic[DatabaseIdT]):
    reason: Literal[CancelReason.history_deleted]
    history_id: DatabaseIdT = Field(..., title="History ID", description="History ID of history that was deleted.")


class GenericInvocationCancellationUserRequest(InvocationMessageBase, Generic[DatabaseIdT]):
    reason: Literal[CancelReason.user_request]


class InvocationFailureMessageBase(InvocationMessageBase, Generic[DatabaseIdT]):
    workflow_step_id: int = Field(
        ..., description="Workflow step id of step that failed.", validation_alias="workflow_step_index"
    )


class GenericInvocationFailureDatasetFailed(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.dataset_failed]
    hda_id: DatabaseIdT = Field(
        ..., title="HistoryDatasetAssociation ID", description="HistoryDatasetAssociation ID that relates to failure."
    )
    dependent_workflow_step_id: Optional[int] = Field(None, description="Workflow step id of step that caused failure.")


class GenericInvocationFailureCollectionFailed(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.collection_failed]
    hdca_id: DatabaseIdT = Field(
        ...,
        title="HistoryDatasetCollectionAssociation ID",
        description="HistoryDatasetCollectionAssociation ID that relates to failure.",
    )
    dependent_workflow_step_id: int = Field(
        ...,
        description="Workflow step id of step that caused failure.",
        validation_alias="dependent_workflow_step_index",
    )


class GenericInvocationFailureJobFailed(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.job_failed]
    job_id: DatabaseIdT = Field(..., title="Job ID", description="Job ID that relates to failure.")
    dependent_workflow_step_id: int = Field(
        ...,
        description="Workflow step id of step that caused failure.",
        validation_alias="dependent_workflow_step_index",
    )


class GenericInvocationFailureOutputNotFound(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.output_not_found]
    output_name: str = Field(..., title="Tool or module output name that was referenced but not produced")
    dependent_workflow_step_id: int = Field(
        ...,
        description="Workflow step id of step that caused failure.",
        validation_alias="dependent_workflow_step_index",
    )


class GenericInvocationFailureExpressionEvaluationFailed(
    InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]
):
    reason: Literal[FailureReason.expression_evaluation_failed]
    details: Optional[str] = Field(None, description="May contain details to help troubleshoot this problem.")


class GenericInvocationFailureWhenNotBoolean(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.when_not_boolean]
    details: str = Field(..., description="Contains details to help troubleshoot this problem.")


class GenericInvocationUnexpectedFailure(InvocationMessageBase, Generic[DatabaseIdT]):
    reason: Literal[FailureReason.unexpected_failure]
    details: Optional[str] = Field(None, description="May contains details to help troubleshoot this problem.")
    workflow_step_id: Optional[int] = Field(
        None, description="Workflow step id of step that failed.", validation_alias="workflow_step_index"
    )


class GenericInvocationWarning(InvocationMessageBase, Generic[DatabaseIdT]):
    reason: WarningReason = Field(..., title="Failure Reason", description="Reason for warning")
    workflow_step_id: Optional[int] = Field(
        None, title="Workflow step id of step that caused a warning.", validation_alias="workflow_step_index"
    )


class GenericInvocationEvaluationWarningWorkflowOutputNotFound(
    GenericInvocationWarning[DatabaseIdT], Generic[DatabaseIdT]
):
    reason: Literal[WarningReason.workflow_output_not_found]
    workflow_step_id: int = Field(
        ..., title="Workflow step id of step that caused a warning.", validation_alias="workflow_step_index"
    )
    output_name: str = Field(
        ..., description="Output that was designated as workflow output but that has not been found"
    )


class GenericInvocationFailureWorkflowParameterInvalid(InvocationFailureMessageBase[DatabaseIdT], Generic[DatabaseIdT]):
    reason: Literal[FailureReason.workflow_parameter_invalid]
    workflow_step_id: int = Field(
        ..., title="Workflow parameter step that failed validation", validation_alias="workflow_step_index"
    )
    details: str = Field(..., description="Message raised by validator")


InvocationCancellationReviewFailed = GenericInvocationCancellationReviewFailed[int]
InvocationCancellationHistoryDeleted = GenericInvocationCancellationHistoryDeleted[int]
InvocationCancellationUserRequest = GenericInvocationCancellationUserRequest[int]
InvocationFailureDatasetFailed = GenericInvocationFailureDatasetFailed[int]
InvocationFailureCollectionFailed = GenericInvocationFailureCollectionFailed[int]
InvocationFailureJobFailed = GenericInvocationFailureJobFailed[int]
InvocationFailureOutputNotFound = GenericInvocationFailureOutputNotFound[int]
InvocationFailureExpressionEvaluationFailed = GenericInvocationFailureExpressionEvaluationFailed[int]
InvocationFailureWhenNotBoolean = GenericInvocationFailureWhenNotBoolean[int]
InvocationUnexpectedFailure = GenericInvocationUnexpectedFailure[int]
InvocationWarningWorkflowOutputNotFound = GenericInvocationEvaluationWarningWorkflowOutputNotFound[int]
InvocationFailureWorkflowParameterInvalid = GenericInvocationFailureWorkflowParameterInvalid[int]

InvocationMessageUnion = Union[
    InvocationCancellationReviewFailed,
    InvocationCancellationHistoryDeleted,
    InvocationCancellationUserRequest,
    InvocationFailureDatasetFailed,
    InvocationFailureCollectionFailed,
    InvocationFailureJobFailed,
    InvocationFailureOutputNotFound,
    InvocationFailureExpressionEvaluationFailed,
    InvocationFailureWhenNotBoolean,
    InvocationUnexpectedFailure,
    InvocationWarningWorkflowOutputNotFound,
    InvocationFailureWorkflowParameterInvalid,
]

InvocationCancellationReviewFailedResponseModel = GenericInvocationCancellationReviewFailed[EncodedDatabaseIdField]
InvocationCancellationHistoryDeletedResponseModel = GenericInvocationCancellationHistoryDeleted[EncodedDatabaseIdField]
InvocationCancellationUserRequestResponseModel = GenericInvocationCancellationUserRequest[EncodedDatabaseIdField]
InvocationFailureDatasetFailedResponseModel = GenericInvocationFailureDatasetFailed[EncodedDatabaseIdField]
InvocationFailureCollectionFailedResponseModel = GenericInvocationFailureCollectionFailed[EncodedDatabaseIdField]
InvocationFailureJobFailedResponseModel = GenericInvocationFailureJobFailed[EncodedDatabaseIdField]
InvocationFailureOutputNotFoundResponseModel = GenericInvocationFailureOutputNotFound[EncodedDatabaseIdField]
InvocationFailureExpressionEvaluationFailedResponseModel = GenericInvocationFailureExpressionEvaluationFailed[
    EncodedDatabaseIdField
]
InvocationFailureWhenNotBooleanResponseModel = GenericInvocationFailureWhenNotBoolean[EncodedDatabaseIdField]
InvocationUnexpectedFailureResponseModel = GenericInvocationUnexpectedFailure[EncodedDatabaseIdField]
InvocationWarningWorkflowOutputNotFoundResponseModel = GenericInvocationEvaluationWarningWorkflowOutputNotFound[
    EncodedDatabaseIdField
]
InvocationFailureWorkflowParameterInvalidResponseModel = GenericInvocationFailureWorkflowParameterInvalid[
    EncodedDatabaseIdField
]

_InvocationMessageResponseUnion = Annotated[
    Union[
        InvocationCancellationReviewFailedResponseModel,
        InvocationCancellationHistoryDeletedResponseModel,
        InvocationCancellationUserRequestResponseModel,
        InvocationFailureDatasetFailedResponseModel,
        InvocationFailureCollectionFailedResponseModel,
        InvocationFailureJobFailedResponseModel,
        InvocationFailureOutputNotFoundResponseModel,
        InvocationFailureExpressionEvaluationFailedResponseModel,
        InvocationFailureWhenNotBooleanResponseModel,
        InvocationUnexpectedFailureResponseModel,
        InvocationWarningWorkflowOutputNotFoundResponseModel,
        InvocationFailureWorkflowParameterInvalidResponseModel,
    ],
    Field(discriminator="reason"),
]

InvocationMessageResponseUnion = TypeAliasType("InvocationMessageResponseUnion", _InvocationMessageResponseUnion)


class InvocationMessageResponseModel(RootModel):
    root: InvocationMessageResponseUnion
    model_config = ConfigDict(from_attributes=True)


class InvocationState(str, Enum):
    NEW = "new"  # Brand new workflow invocation... maybe this should be same as READY
    REQUIRES_MATERIALIZATION = "requires_materialization"  # an otherwise NEW or READY workflow that requires inputs to be materialized (undeferred)
    READY = "ready"  # Workflow ready for another iteration of scheduling.
    SCHEDULED = "scheduled"  # Workflow has been scheduled.
    CANCELLED = "cancelled"
    CANCELLING = "cancelling"  # invocation scheduler will cancel job in next iteration.
    FAILED = "failed"


class InvocationStepState(str, Enum):
    NEW = "new"  # Brand new workflow invocation step
    READY = "ready"  # Workflow invocation step ready for another iteration of scheduling.
    SCHEDULED = "scheduled"  # Workflow invocation step has been scheduled.
    # CANCELLED = 'cancelled',  TODO: implement and expose
    # FAILED = 'failed',  TODO: implement and expose


class InvocationStepOutput(Model):
    src: INVOCATION_STEP_OUTPUT_SRC = Field(
        literal_to_value(INVOCATION_STEP_OUTPUT_SRC),
        title="src",
        description="The source model of the output.",
    )
    id: EncodedDatabaseIdField = Field(
        ...,
        title="Dataset ID",
        description="Dataset ID of the workflow step output.",
    )
    uuid: Optional[UUID4] = Field(
        None,
        title="UUID",
        description="Universal unique identifier of the workflow step output dataset.",
    )


class InvocationStepCollectionOutput(Model):
    src: INVOCATION_STEP_COLLECTION_OUTPUT_SRC = Field(
        literal_to_value(INVOCATION_STEP_COLLECTION_OUTPUT_SRC),
        title="src",
        description="The source model of the output.",
    )
    id: EncodedDatabaseIdField = Field(
        ...,
        title="Dataset Collection ID",
        description="Dataset Collection ID of the workflow step output.",
    )


class InvocationStep(Model, WithModelClass):
    """Information about workflow invocation step"""

    model_class: INVOCATION_STEP_MODEL_CLASS = ModelClassField(INVOCATION_STEP_MODEL_CLASS)
    id: Annotated[EncodedDatabaseIdField, Field(..., title="Invocation Step ID")]
    update_time: Optional[datetime] = schema.UpdateTimeField
    job_id: Annotated[
        Optional[EncodedDatabaseIdField],
        Field(
            default=None,
            title="Job ID",
            description="The encoded ID of the job associated with this workflow invocation step.",
        ),
    ]
    workflow_step_id: Annotated[
        EncodedDatabaseIdField,
        Field(
            ...,
            title="Workflow step ID",
            description="The encoded ID of the workflow step associated with this workflow invocation step.",
        ),
    ]
    subworkflow_invocation_id: Annotated[
        Optional[EncodedDatabaseIdField],
        Field(
            default=None,
            title="Subworkflow invocation ID",
            description="The encoded ID of the subworkflow invocation.",
        ),
    ]
    state: Optional[Union[InvocationStepState, JobState]] = Field(
        default=None,
        title="State of the invocation step",
        description="Describes where in the scheduling process the workflow invocation step is.",
    )
    action: Optional[bool] = InvocationStepActionField
    order_index: int = Field(
        ...,
        title="Order index",
        description="The index of the workflow step in the workflow.",
    )
    workflow_step_label: Optional[str] = Field(
        default=None,
        title="Step label",
        description="The label of the workflow step",
    )
    workflow_step_uuid: Optional[UUID4] = Field(
        None,
        title="UUID",
        description="Universal unique identifier of the workflow step.",
    )
    outputs: dict[str, InvocationStepOutput] = Field(
        {},
        title="Outputs",
        description="The outputs of the workflow invocation step.",
    )
    output_collections: dict[str, InvocationStepCollectionOutput] = Field(
        {},
        title="Output collections",
        description="The dataset collection outputs of the workflow invocation step.",
    )
    jobs: list[schema.JobBaseModel] = Field(
        [],
        title="Jobs",
        description="Jobs associated with the workflow invocation step.",
    )
    implicit_collection_jobs_id: Optional[EncodedDatabaseIdField] = Field(
        None,
        title="Implicit Collection Jobs ID",
        description="The implicit collection job ID associated with the workflow invocation step.",
    )


class InvocationReport(Model, WithModelClass):
    """Report describing workflow invocation"""

    render_format: REPORT_RENDER_FORMAT_MARKDOWN = Field(
        literal_to_value(REPORT_RENDER_FORMAT_MARKDOWN),
        title="Render format",
        description="Format of the invocation report.",
    )
    markdown: Optional[str] = Field(
        default=None,
        title="Markdown",
        description="Raw galaxy-flavored markdown contents of the report.",
    )
    invocation_markdown: Optional[str] = Field(
        default=None,
        title="Markdown",
        description="Raw galaxy-flavored markdown contents of the report.",
    )
    model_class: schema.INVOCATION_REPORT_MODEL_CLASS = ModelClassField(schema.INVOCATION_REPORT_MODEL_CLASS)
    id: EncodedDatabaseIdField = Field(
        ...,
        title="Workflow ID",
        description="The workflow this invocation has been triggered for.",
    )
    username: str = Field(
        ...,
        title="Username",
        description="The name of the user who owns this report.",
    )
    title: str = Field(
        ...,
        title="Title",
        description="The name of the report.",
    )
    generate_time: Optional[str] = schema.GenerateTimeField
    generate_version: Optional[str] = schema.GenerateVersionField

    errors: Optional[list[dict[str, Any]]] = Field(
        default=None,
        title="Errors",
        description="Errors associated with the invocation.",
    )

    history_datasets: Optional[dict[str, Any]] = Field(
        default=None,
        title="History datasets",
        description="History datasets associated with the invocation.",
    )
    workflows: Optional[dict[str, Any]] = Field(
        default=None,
        title="Workflows",
        description="Workflows associated with the invocation.",
    )
    history_dataset_collections: Optional[dict[str, Any]] = Field(
        default=None,
        title="History dataset collections",
        description="History dataset collections associated with the invocation.",
    )
    jobs: Optional[dict[str, Any]] = Field(
        default=None,
        title="Jobs",
        description="Jobs associated with the invocation.",
    )
    histories: Optional[dict[str, Any]] = Field(
        default=None,
        title="Histories",
        description="Histories associated with the invocation.",
    )
    invocations: Optional[dict[str, Any]] = Field(
        default=None,
        title="Invocations",
        description="Other invocations associated with the invocation.",
    )


class ReportInvocationErrorPayload(Model):
    invocation_id: DecodedDatabaseIdField = Field(
        default=...,
        title="Invocation ID",
        description="The ID of the invocation related to the error.",
    )
    email: Optional[str] = Field(
        default=None,
        title="Email",
        description="Email address for communication with the user. Only required for anonymous users.",
    )
    message: Optional[str] = Field(
        default=None,
        title="Message",
        description="The optional message sent with the error report.",
    )


class InvocationUpdatePayload(Model):
    action: bool = InvocationStepActionField


class InvocationIOBase(Model):
    id: Optional[EncodedDatabaseIdField] = Field(
        default=None, title="ID", description="The encoded ID of the dataset/dataset collection."
    )
    workflow_step_id: EncodedDatabaseIdField = Field(
        ...,
        title="Workflow step ID",
        description="The encoded ID of the workflow step associated with the dataset/dataset collection.",
    )


class InvocationInput(InvocationIOBase):
    label: Optional[str] = Field(
        default=None,
        title="Label",
        description="Label of the workflow step associated with the input dataset/dataset collection.",
    )
    src: Union[Literal[DataItemSourceType.hda], Literal[DataItemSourceType.hdca]] = Field(
        default=..., title="Source", description="Source type of the input dataset/dataset collection."
    )


class InvocationInputParameter(Model):
    # TODO - Change the type of parameter_value, when all valid types are known
    parameter_value: Any = Field(default=..., title="Parameter value", description="Value of the input parameter.")
    label: str = Field(
        default=..., title="Label", description="Label of the workflow step associated with the input parameter."
    )
    workflow_step_id: EncodedDatabaseIdField = Field(
        default=...,
        title="Workflow step ID",
        description="The encoded ID of the workflow step associated with the input parameter.",
    )


class InvocationOutput(InvocationIOBase):
    src: Literal[DataItemSourceType.hda] = Field(
        default=..., title="Source", description="Source model of the output dataset."
    )


class InvocationOutputCollection(InvocationIOBase):
    src: Literal[DataItemSourceType.hdca] = Field(
        default=..., title="Source", description="Source model of the output dataset collection."
    )


InvocationWorkflowIdField = Field(
    title="Workflow ID", description="The encoded Workflow ID associated with the invocation."
)


class WorkflowInvocationCollectionView(Model, WithModelClass):
    id: EncodedDatabaseIdField = InvocationIdField
    create_time: datetime = CreateTimeField
    update_time: datetime = UpdateTimeField
    workflow_id: EncodedDatabaseIdField = InvocationWorkflowIdField
    history_id: EncodedDatabaseIdField = Field(
        default=...,
        title="History ID",
        description="The encoded ID of the history associated with the invocation.",
    )
    # The uuid version here is 1, which deviates from the other UUIDs used as they are version 4.
    uuid: Optional[Union[UUID4, UUID1]] = Field(
        default=None, title="UUID", description="Universal unique identifier of the workflow invocation."
    )
    state: InvocationState = Field(default=..., title="Invocation state", description="State of workflow invocation.")
    landing_uuid: Optional[UUID4] = Field(
        default=None,
        title="Landing UUID",
        description="The UUID of the workflow landing request associated with this invocation.",
    )
    model_class: INVOCATION_MODEL_CLASS = ModelClassField(INVOCATION_MODEL_CLASS)


class WorkflowInvocationElementView(WorkflowInvocationCollectionView):
    steps: list[InvocationStep] = Field(default=..., title="Steps", description="Steps of the workflow invocation.")
    inputs: dict[str, InvocationInput] = Field(
        default=..., title="Inputs", description="Input datasets/dataset collections of the workflow invocation."
    )
    input_step_parameters: dict[str, InvocationInputParameter] = Field(
        default=..., title="Input step parameters", description="Input step parameters of the workflow invocation."
    )
    outputs: dict[str, InvocationOutput] = Field(
        default=..., title="Outputs", description="Output datasets of the workflow invocation."
    )
    output_collections: dict[str, InvocationOutputCollection] = Field(
        default=...,
        title="Output collections",
        description="Output dataset collections of the workflow invocation.",
    )
    output_values: dict[str, Any] = Field(
        default=..., title="Output values", description="Output values of the workflow invocation."
    )
    messages: list[InvocationMessageResponseUnion] = Field(
        default=...,
        title="Messages",
        description="A list of messages about why the invocation did not succeed.",
    )


class WorkflowInvocationResponse(RootModel):
    root: Annotated[
        Union[WorkflowInvocationElementView, WorkflowInvocationCollectionView], Field(union_mode="left_to_right")
    ]


class WorkflowInvocationRequestModel(Model):
    """Model a workflow invocation request (InvokeWorkflowPayload) for an existing invocation."""

    history_id: str = Field(
        ...,
        title="History ID",
        description="The encoded history id the workflow was run in.",
    )
    workflow_id: str = Field(title="Workflow ID", description="The encoded Workflow ID associated with the invocation.")
    inputs: dict[str, Any] = Field(
        ...,
        title="Inputs",
        description="Values for inputs",
    )
    inputs_by: str = Field(
        ...,
        title="Inputs by",
        description=INPUTS_BY_DESCRIPTION,
    )
    replacement_params: Optional[dict[str, Any]] = ReplacementParametersField
    resource_params: Optional[dict[str, Any]] = ResourceParametersField
    use_cached_job: bool = UseCachedJobField
    preferred_object_store_id: Optional[str] = PreferredObjectStoreIdField
    preferred_intermediate_object_store_id: Optional[str] = PreferredIntermediateObjectStoreIdField
    preferred_outputs_object_store_id: Optional[str] = PreferredOutputsObjectStoreIdField
    parameters_normalized: Literal[True] = Field(
        True,
        title=STEP_PARAMETERS_NORMALIZED_TITLE,
        description=STEP_PARAMETERS_NORMALIZED_DESCRIPTION,
    )
    parameters: Optional[dict[str, Any]] = Field(
        None,
        title=STEP_PARAMETERS_TITLE,
        description=f"{STEP_PARAMETERS_DESCRIPTION} If these are set, the workflow was not executed in a best-practice fashion and we the resulting invocation request may not fully reflect the executed workflow state.",
    )
    instance: Literal[True] = Field(
        True,
        title="Is instance",
        description="This API yields a particular workflow instance, newer workflows belonging to the same storedworkflow may have different state.",
    )


class InvocationJobsSummaryBaseModel(Model):
    id: EncodedDatabaseIdField = InvocationIdField
    states: dict[JobState, int] = Field(
        default=..., title="States", description="The states of all the jobs related to the Invocation."
    )
    populated_state: JobState = Field(
        default=...,
        title="Populated state",
        description="The absolute state of all the jobs related to the Invocation.",
    )


class InvocationJobsResponse(InvocationJobsSummaryBaseModel):
    model: INVOCATION_MODEL_CLASS


class InvocationStepJobsResponseStepModel(InvocationJobsSummaryBaseModel):
    model: INVOCATION_STEP_MODEL_CLASS


class InvocationStepJobsResponseJobModel(InvocationJobsSummaryBaseModel):
    model: JOB_MODEL_CLASS
    id: EncodedDatabaseIdField = Field(
        default=...,
        title="ID",
        description="The encoded ID of the job.",
    )


class InvocationStepJobsResponseCollectionJobsModel(InvocationJobsSummaryBaseModel):
    model: IMPLICIT_COLLECTION_JOBS_MODEL_CLASS
    id: EncodedDatabaseIdField = Field(
        default=...,
        title="ID",
        description="The encoded ID of the collection job.",
    )


class CreateInvocationFromStore(StoreContentSource):
    history_id: int = Field(
        default=..., title="History ID", description="The ID of the history associated with the invocations."
    )


class InvocationSerializationView(str, Enum):
    element = "element"
    collection = "collection"


class InvocationSerializationParams(BaseModel):
    """Contains common parameters for customizing model serialization."""

    view: Optional[InvocationSerializationView] = Field(
        default=None,
        title="View",
        description=(
            "The name of the view used to serialize this item. "
            "This will return a predefined set of attributes of the item."
        ),
        examples=["element"],
    )
    step_details: bool = Field(
        default=False,
        title="Include step details",
        description="Include details for individual invocation steps and populate a steps attribute in the resulting dictionary",
    )
    legacy_job_state: bool = Field(
        default=False,
        description="""Populate the invocation step state with the job state instead of the invocation step state.
        This will also produce one step per job in mapping jobs to mimic the older behavior with respect to collections.
        Partially scheduled steps may provide incomplete information and the listed steps outputs
        are not the mapped over step outputs but the individual job outputs.""",
        # TODO: also deprecate on python side, https://github.com/pydantic/pydantic/issues/2255
        json_schema_extra={"deprecated": True},
    )


class CreateInvocationsFromStorePayload(CreateInvocationFromStore, InvocationSerializationParams):
    history_id: DecodedDatabaseIdField = Field(
        default=...,
        title="History ID",
        description="The ID of the history associated with the invocations.",
    )



import json
from enum import Enum
from typing import (
    Annotated,
    Any,
    Optional,
    Union,
)

from pydantic import (
    ConfigDict,
    Field,
    RootModel,
)
from pydantic.functional_validators import field_validator
from typing_extensions import (
    Literal,
)

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
    EncodedLibraryFolderDatabaseIdField,
    LibraryFolderDatabaseIdField,
    ModelClassField,
)
from galaxy.schema.schema import (
    Model,
    TagCollection,
)


class UploadOption(str, Enum):
    upload_file = "upload_file"
    upload_paths = "upload_paths"
    upload_directory = "upload_directory"


class CreateType(str, Enum):
    file = "file"
    folder = "folder"
    collection = "collection"


class LinkDataOnly(str, Enum):
    copy_files = "copy_files"
    link_to_files = "link_to_files"


class LibraryContentsCreatePayload(Model):
    create_type: CreateType = Field(
        ...,
        description="the type of item to create",
    )
    upload_option: UploadOption = Field(
        UploadOption.upload_file,
        description="the method to use for uploading files",
        deprecated=True,
    )
    folder_id: LibraryFolderDatabaseIdField = Field(
        ...,
        description="the encoded id of the parent folder of the new item",
    )
    tag_using_filenames: bool = Field(
        False,
        description="create tags on datasets using the file's original name",
    )
    tags: list[str] = Field(
        [],
        description="create the given list of tags on datasets",
    )
    from_hda_id: Optional[DecodedDatabaseIdField] = Field(
        None,
        description="(only if create_type is 'file') the encoded id of an accessible HDA to copy into the library",
    )
    from_hdca_id: Optional[DecodedDatabaseIdField] = Field(
        None,
        description="(only if create_type is 'file') the encoded id of an accessible HDCA to copy into the library",
    )
    ldda_message: str = Field(
        "",
        description="the new message attribute of the LDDA created",
    )
    extended_metadata: Optional[dict[str, Any]] = Field(
        None,
        description="sub-dictionary containing any extended metadata to associate with the item",
    )

    @field_validator("tags", mode="before", check_fields=False)
    @classmethod
    def tags_string_to_json(cls, v):
        if isinstance(v, str):
            return json.loads(v)
        return v


class LibraryContentsFileCreatePayload(LibraryContentsCreatePayload):
    dbkey: Union[str, list] = Field(
        "?",
        title="database key",
    )
    roles: str = Field(
        "",
        title="user selected roles",
    )
    file_type: Optional[str] = Field(
        None,
        title="file type",
    )
    server_dir: str = Field(
        "",
        description="(only if upload_option is 'upload_directory') relative path of the "
        "subdirectory of Galaxy ``library_import_dir`` (if admin) or "
        "``user_library_import_dir`` (if non-admin) to upload. "
        "All and only the files (i.e. no subdirectories) contained "
        "in the specified directory will be uploaded.",
    )
    filesystem_paths: str = Field(
        "",
        description="(only if upload_option is 'upload_paths' and the user is an admin) "
        "file paths on the Galaxy server to upload to the library, one file per line",
    )
    link_data_only: LinkDataOnly = Field(
        LinkDataOnly.copy_files,
        description="(only when upload_option is 'upload_directory' or 'upload_paths')."
        "Setting to 'link_to_files' symlinks instead of copying the files",
    )
    uuid: Optional[str] = Field(
        None,
        title="UUID of the dataset to upload",
    )
    upload_files: Optional[list[dict[str, Any]]] = Field(
        None,
        title="list of the uploaded files",
    )

    # uploaded file fields
    model_config = ConfigDict(extra="allow")


class LibraryContentsFolderCreatePayload(LibraryContentsCreatePayload):
    name: str = Field(
        "",
        title="name of the folder to create",
    )
    description: str = Field(
        "",
        title="description of the folder to create",
    )


class LibraryContentsCollectionCreatePayload(LibraryContentsCreatePayload):
    collection_type: str = Field(
        ...,
        title="the type of collection to create",
    )
    element_identifiers: list[dict[str, Any]] = Field(
        ...,
        title="list of dictionaries containing the element identifiers for the collection",
    )
    name: Optional[str] = Field(
        None,
        title="the name of the collection",
    )
    hide_source_items: bool = Field(
        False,
        description="if True, hide the source items in the collection",
    )
    copy_elements: bool = Field(
        False,
        description="if True, copy the elements into the collection",
    )


class LibraryContentsUpdatePayload(Model):
    converted_dataset_id: Optional[DecodedDatabaseIdField] = Field(
        None,
        title="the decoded id of the dataset",
    )


class LibraryContentsDeletePayload(Model):
    purge: bool = Field(
        False,
        description="if True, purge the library dataset",
    )


class LibraryContentsIndexResponse(Model):
    type: str
    name: str
    url: str


class LibraryContentsIndexFolderResponse(LibraryContentsIndexResponse):
    id: EncodedLibraryFolderDatabaseIdField


class LibraryContentsIndexDatasetResponse(LibraryContentsIndexResponse):
    id: EncodedDatabaseIdField


class LibraryContentsIndexListResponse(RootModel):
    root: list[Union[LibraryContentsIndexFolderResponse, LibraryContentsIndexDatasetResponse]]


class LibraryContentsShowResponse(Model):
    name: str
    genome_build: Optional[str]
    update_time: str
    parent_library_id: EncodedDatabaseIdField


class LibraryContentsShowFolderResponse(LibraryContentsShowResponse):
    model_class: Annotated[Literal["LibraryFolder"], ModelClassField(Literal["LibraryFolder"])]
    id: EncodedLibraryFolderDatabaseIdField
    parent_id: Optional[EncodedLibraryFolderDatabaseIdField]
    description: str
    item_count: int
    deleted: bool
    library_path: list[str]


class LibraryContentsShowDatasetResponse(LibraryContentsShowResponse):
    model_class: Annotated[Literal["LibraryDataset"], ModelClassField(Literal["LibraryDataset"])]
    id: EncodedDatabaseIdField
    ldda_id: EncodedDatabaseIdField
    folder_id: EncodedLibraryFolderDatabaseIdField
    state: str
    file_name: str
    created_from_basename: Optional[str]
    uploaded_by: Optional[str]
    message: Optional[str]
    date_uploaded: str
    file_size: int
    file_ext: str
    data_type: str
    misc_info: Optional[str]
    misc_blurb: Optional[str]
    peek: Optional[str]
    uuid: str
    tags: TagCollection

    # metadata fields
    model_config = ConfigDict(extra="allow")


class LibraryContentsCreateResponse(Model):
    name: str
    url: str


class LibraryContentsCreateFolderResponse(LibraryContentsCreateResponse):
    id: EncodedLibraryFolderDatabaseIdField


class LibraryContentsCreateFileResponse(LibraryContentsCreateResponse):
    id: EncodedDatabaseIdField


class LibraryContentsCreateFolderListResponse(RootModel):
    root: list[LibraryContentsCreateFolderResponse]


class LibraryContentsCreateFileListResponse(RootModel):
    root: list[LibraryContentsCreateFileResponse]


class LibraryContentsCreateDatasetResponse(Model):
    # id, library_dataset_id, parent_library_id should change to EncodedDatabaseIdField latter
    # because they are encoded ids in _copy_hda_to_library_folder and _copy_hdca_to_library_folder
    # functions that are shared by LibraryFolderContentsService too
    id: str
    hda_ldda: str
    model_class: Annotated[
        Literal["LibraryDatasetDatasetAssociation"], ModelClassField(Literal["LibraryDatasetDatasetAssociation"])
    ]
    name: str
    deleted: bool
    visible: bool
    state: str
    library_dataset_id: str
    file_size: int
    file_name: str
    update_time: str
    file_ext: str
    data_type: str
    genome_build: str
    misc_info: Optional[str]
    misc_blurb: Optional[str]
    created_from_basename: Optional[str]
    uuid: str
    parent_library_id: str

    # metadata fields
    model_config = ConfigDict(extra="allow")


class LibraryContentsCreateDatasetCollectionResponse(RootModel):
    root: list[LibraryContentsCreateDatasetResponse]


class LibraryContentsDeleteResponse(Model):
    id: EncodedDatabaseIdField
    deleted: bool


class LibraryContentsPurgedResponse(LibraryContentsDeleteResponse):
    purged: bool


AnyLibraryContentsShowResponse = Union[
    LibraryContentsShowFolderResponse,
    LibraryContentsShowDatasetResponse,
]

AnyLibraryContentsCreatePayload = Union[
    LibraryContentsFolderCreatePayload, LibraryContentsFileCreatePayload, LibraryContentsCollectionCreatePayload
]

AnyLibraryContentsCreateResponse = Union[
    LibraryContentsCreateFolderListResponse,
    LibraryContentsCreateFileListResponse,
    LibraryContentsCreateDatasetCollectionResponse,
    LibraryContentsCreateDatasetResponse,
]



# generated by datamodel-codegen:
#   filename:  https://opensource.ieee.org/2791-object/ieee-2791-schema/-/raw/master/parametric_domain.json
#   timestamp: 2022-09-13T23:51:53+00:00

from __future__ import annotations

from typing import List

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
)


class ParametricDomainItem(BaseModel):
    model_config = ConfigDict(extra="forbid")

    param: str = Field(
        ..., description="Specific variables for the computational workflow", examples=["seed"], title="param"
    )
    value: str = Field(
        ...,
        description="Specific (non-default) parameter values for the computational workflow",
        examples=["14"],
        title="value",
    )
    step: str = Field(
        ...,
        description="Refers to the specific step of the workflow relevant to the parameters specified in 'param' and 'value'",
        examples=["1"],
        pattern="^(.*)$",
        title="step",
    )


class ParametricDomain(RootModel):
    root: List[ParametricDomainItem] = Field(
        ...,
        description="This represents the list of NON-default parameters customizing the computational flow which can affect the output of the calculations. These fields can be custom to each kind of analysis and are tied to a particular pipeline implementation",
        title="Parametric Domain",
    )



import yaml

from galaxy.util.resources import resource_string


def load_terms_raw():
    new_data_yaml = resource_string(__name__, "terms.yml")
    terms_raw = yaml.safe_load(new_data_yaml)
    return terms_raw


class HelpTerms:
    def __init__(self):
        self.terms = load_terms_raw()

    def get_term(self, term_path: str) -> str:
        root = self.terms
        for part in term_path.split("."):
            if part not in root:
                raise KeyError(f"Term '{term_path}' not found.")
            root = root[part]
        assert isinstance(root, str), f"Term '{term_path}' is not a string."
        return root



from enum import Enum
from typing import (
    Optional,
    Union,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
)


class Requirement(str, Enum):
    """Available types of job sources (model classes) that produce dataset collections."""

    LOGGED_IN = "logged_in"
    NEW_HISTORY = "new_history"
    ADMIN = "admin"


class TourCore(BaseModel):
    name: str = Field(title="Name", description="Name of tour")
    description: str = Field(title="Description", description="Tour description")
    tags: list[str] = Field(title="Tags", description="Topic topic tags")
    requirements: list[Requirement] = Field(title="Requirements", description="Requirements to run the tour.")
    model_config = ConfigDict(use_enum_values=True)


class Tour(TourCore):
    id: str = Field(title="Identifier", description="Tour identifier")


class TourList(RootModel):
    root: list[Tour] = Field(title="List of tours", default=[])


class TourStep(BaseModel):
    title: Optional[str] = Field(None, title="Title", description="Title displayed in the header of the step container")
    content: Optional[str] = Field(None, title="Content", description="Text shown to the user")
    element: Optional[str] = Field(
        None, title="Element", description="CSS selector for the element to be described/clicked"
    )
    placement: Optional[str] = Field(
        None, title="Placement", description="Placement of the text box relative to the selected element"
    )
    preclick: Optional[Union[bool, list[str]]] = Field(
        None, title="Pre-click", description="Elements that receive a click() event before the step is shown"
    )
    postclick: Optional[Union[bool, list[str]]] = Field(
        None, title="Post-click", description="Elements that receive a click() event after the step is shown"
    )
    textinsert: Optional[str] = Field(
        None, title="Text-insert", description="Text to insert if element is a text box (e.g. tool search or upload)"
    )
    orphan: Optional[bool] = Field(None, title="Orphan", description="If true, the step is an orphan step")


class TourDetails(TourCore):
    title_default: Optional[str] = Field(None, title="Default title", description="Default title for each step")
    steps: list[TourStep] = Field(title="Steps", description="Tour steps")



from datetime import datetime
from enum import Enum
from typing import (
    Union,
)

from pydantic import Field
from typing_extensions import Literal

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
)
from galaxy.schema.schema import (
    Model,
    UpdateTimeField,
)


class CleanableItemsSummary(Model):
    total_size: int = Field(
        ...,
        title="Total Size",
        description="The total size in bytes that can be recovered by purging all the items.",
    )
    total_items: int = Field(
        ...,
        title="Total Items",
        description="The total number of items that could be purged.",
    )


StoredItemType = Union[Literal["history"], Literal["dataset"]]


class StoredItem(Model):
    id: EncodedDatabaseIdField
    name: str
    type: StoredItemType
    size: int
    update_time: datetime = UpdateTimeField


class StoredItemOrderBy(str, Enum):
    """Available options for sorting Stored Items results."""

    NAME_ASC = "name-asc"
    NAME_DSC = "name-dsc"
    SIZE_ASC = "size-asc"
    SIZE_DSC = "size-dsc"
    UPDATE_TIME_ASC = "update_time-asc"
    UPDATE_TIME_DSC = "update_time-dsc"


class StorageItemCleanupError(Model):
    item_id: EncodedDatabaseIdField
    error: str


class CleanupStorageItemsRequest(Model):
    item_ids: list[DecodedDatabaseIdField]


class StorageItemsCleanupResult(Model):
    total_item_count: int
    success_item_count: int
    total_free_bytes: int
    errors: list[StorageItemCleanupError]



from datetime import datetime
from typing import (
    Annotated,
    Literal,
    Optional,
)

from pydantic import (
    Field,
    RootModel,
)

from galaxy.schema.fields import (
    DecodedDatabaseIdField,
    EncodedDatabaseIdField,
)
from galaxy.schema.schema import Model

SOURCE_TYPE = Literal["tool"]


class CredentialResponse(Model):
    name: Annotated[
        str,
        Field(
            description="The name of the credential.",
        ),
    ]


class VariableResponse(CredentialResponse):
    value: Annotated[
        Optional[str],
        Field(
            None,
            description="The value of the variable (for variables, not secrets).",
        ),
    ]


class SecretResponse(CredentialResponse):
    is_set: Annotated[
        bool,
        Field(
            description="Whether the secret has been set (value is not exposed).",
        ),
    ]


class ServiceCredentialGroupResponse(Model):
    id: Annotated[
        EncodedDatabaseIdField,
        Field(
            description="Encoded ID of the credential group.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the credential group.",
        ),
    ]
    variables: list[VariableResponse]
    secrets: list[SecretResponse]
    update_time: Annotated[
        datetime,
        Field(
            description="The last time the credential group was updated.",
        ),
    ]


class ServiceParameterDefinition(Model):
    name: Annotated[
        str,
        Field(
            description="The name of the credential definition.",
        ),
    ]
    label: Annotated[
        str,
        Field(
            description="The human-readable label for the credential.",
        ),
    ]
    description: Annotated[
        str,
        Field(
            description="A description of what this credential is used for.",
        ),
    ]
    optional: Annotated[
        bool,
        Field(
            description="Whether this credential is optional or required.",
        ),
    ]


class UserServiceCredentialsResponse(Model):
    user_id: Annotated[
        EncodedDatabaseIdField,
        Field(
            description="The ID of the user who owns these credentials.",
        ),
    ]
    id: Annotated[
        EncodedDatabaseIdField,
        Field(
            description="The encoded ID of the user credentials.",
        ),
    ]
    source_type: Annotated[
        SOURCE_TYPE,
        Field(
            description="The type of source (e.g., 'tool').",
        ),
    ]
    source_id: Annotated[
        str,
        Field(
            description="The ID of the source (e.g., tool ID).",
        ),
    ]
    source_version: Annotated[
        str,
        Field(
            description="The version of the source.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the service requiring credentials.",
        ),
    ]
    version: Annotated[
        str,
        Field(
            description="The version of the service.",
        ),
    ]
    current_group_id: Annotated[
        Optional[EncodedDatabaseIdField],
        Field(
            None,
            description="The ID of the currently active credential group.",
        ),
    ]
    groups: list[ServiceCredentialGroupResponse]


class UserServiceCredentialsListResponse(RootModel):
    root: list[UserServiceCredentialsResponse]


class ServiceCredentialsDefinition(Model):
    name: Annotated[
        str,
        Field(
            description="The name of the service.",
        ),
    ]
    version: Annotated[
        str,
        Field(
            description="The version of the service.",
        ),
    ]
    description: Annotated[
        str,
        Field(
            description="A description of the service.",
        ),
    ]
    label: Annotated[
        Optional[str],
        Field(
            None,
            description="A human-readable label for the service.",
        ),
    ]
    optional: Annotated[
        bool,
        Field(
            description="If true, tools can run without credentials; if false, credentials must be provided before execution.",
        ),
    ]
    variables: list[ServiceParameterDefinition]
    secrets: list[ServiceParameterDefinition]


class UserServiceCredentialsWithDefinitionResponse(UserServiceCredentialsResponse):
    definition: ServiceCredentialsDefinition


class ExtendedUserCredentialsListResponse(RootModel):
    root: list[UserServiceCredentialsWithDefinitionResponse]


class CredentialPayload(Model):
    name: Annotated[
        str,
        Field(
            title="Credential Name",
            description="The name of the credential (variable or secret).",
        ),
    ]
    value: Annotated[
        Optional[str],
        Field(
            None,
            description="The value of the credential.",
        ),
    ]


class ServiceCredentialGroupPayload(Model):
    name: Annotated[
        str,
        Field(
            min_length=3,
            description="The name of the credential group (minimum 3 characters).",
        ),
    ]
    variables: Annotated[
        list[CredentialPayload],
        Field(
            description="List of variables for this credential group.",
        ),
    ]
    secrets: Annotated[
        list[CredentialPayload],
        Field(
            description="List of secrets for this credential group.",
        ),
    ]


class SourceCredentialPayload(Model):
    source_type: Annotated[
        SOURCE_TYPE,
        Field(
            description="The type of source requiring credentials.",
        ),
    ]
    source_id: Annotated[
        str,
        Field(
            description="The ID of the source (e.g., tool ID).",
        ),
    ]
    source_version: Annotated[
        str,
        Field(
            description="The version of the source.",
        ),
    ]


class ServiceCredentialPayload(Model):
    name: Annotated[
        str,
        Field(
            description="The name of the service requiring credentials.",
        ),
    ]
    version: Annotated[
        str,
        Field(
            description="The version of the service.",
        ),
    ]
    group: Annotated[
        ServiceCredentialGroupPayload,
        Field(
            description="The credential group containing variables and secrets.",
        ),
    ]


class CreateSourceCredentialsPayload(SourceCredentialPayload):
    service_credential: Annotated[
        ServiceCredentialPayload,
        Field(
            description="The service credential details including group and credentials.",
        ),
    ]


class SelectCurrentGroupPayload(Model):
    user_credentials_id: Annotated[
        DecodedDatabaseIdField,
        Field(
            description="The ID of the user credentials to update.",
        ),
    ]
    current_group_id: Annotated[
        Optional[DecodedDatabaseIdField],
        Field(
            None,
            description="The ID of the group to set as current (None to unset).",
        ),
    ]


class SelectServiceCredentialPayload(SourceCredentialPayload):
    service_credentials: Annotated[
        list[SelectCurrentGroupPayload],
        Field(
            description="List of user credentials to update with current group selections.",
        ),
    ]


class SelectedGroup(Model):
    id: Annotated[
        DecodedDatabaseIdField,
        Field(
            description="The ID of the selected credential group.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the selected credential group.",
        ),
    ]


class SelectedGroupResponse(Model):
    id: Annotated[
        Optional[EncodedDatabaseIdField],
        Field(
            description="The encoded ID of the user credentials. If null, the group has been deleted by the user.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the selected credential group.",
        ),
    ]


class ServiceCredentialsContext(Model):
    user_credentials_id: Annotated[
        DecodedDatabaseIdField,
        Field(
            description="The ID of the user credentials.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the service.",
        ),
    ]
    version: Annotated[
        str,
        Field(
            description="The version of the service.",
        ),
    ]
    selected_group: Annotated[
        SelectedGroup,
        Field(
            ...,
            description="The currently selected credential group.",
        ),
    ]


class ServiceCredentialsContextResponse(Model):
    user_credentials_id: Annotated[
        Optional[EncodedDatabaseIdField],
        Field(
            description="The encoded ID of the user credentials. If null, the credentials have been deleted by the user.",
        ),
    ]
    name: Annotated[
        str,
        Field(
            description="The name of the service.",
        ),
    ]
    version: Annotated[
        str,
        Field(
            description="The version of the service.",
        ),
    ]
    selected_group: Annotated[
        SelectedGroupResponse,
        Field(
            ...,
            description="The currently selected credential group.",
        ),
    ]


class CredentialsContext(RootModel):
    """Context for credentials to be used during tool execution.

    Contains the list of selected service credentials provided by the user.
    """

    root: list[ServiceCredentialsContext]


class CredentialsContextResponse(RootModel):
    """Response model for credentials context used during tool execution.

    Contains the list of selected service credentials provided by the user.
    """

    root: list[ServiceCredentialsContextResponse]
